import java.util.*;
/*
 * Based on HerongYang.com's work
 */
public class Philosopher extends Thread {
	public static final int numberOfThreads = 5;
	public static Object[] listOfLocks = new Object[numberOfThreads];
	public static String[] dinerTable = new String[4 * numberOfThreads];
	public static String[] lockedDiner = new String[4 * numberOfThreads];
	public static Random randomGenerator = new Random();
	public static int unitOfTime = 500;
	private int threadIndex;
	private static String array2String(Object arr[], String delimiter) {
		StringBuilder sb = new StringBuilder();
		for (Object obj : arr)
			sb.append(obj.toString()).append(delimiter);
		return sb.substring(0, sb.length() - 1);
	}
	private static String getInfo(String s[]) {
		String result = "";
		for (int i = 0; i < numberOfThreads; i++) {
			if (s[4 * i].equals("^")) {
				result += "Fork:Idle  ";
			} else {
				result += "Fork:Taken ";
			}
			if (s[4 * i + 2].equals("_O_")) {
				result += "[P" + (i + 1) + ":No fork] ";
			} else if (s[4 * i + 2].equals("<O>")) {
				result += "[P" + (i + 1) + ":2 forks] ";
			} else {
				result += "[P" + (i + 1) + ": 1 fork] ";
			}
		}
		return result;
	}
	public Philosopher(int i) {
		threadIndex = i;
	}
	public void run() {
		while (!isInterrupted()) {
			try {
				sleep(unitOfTime * randomGenerator.nextInt(6));
			} catch (InterruptedException e) {
				break;
			}
			// Try to get the fork on the left
			Object leftLock = listOfLocks[threadIndex];
			synchronized (leftLock) {
				int i = 4 * threadIndex;
				dinerTable[i] = " ";
				dinerTable[i + 1] = "^";
				dinerTable[i + 2] = "<O ";
				try {
					sleep(unitOfTime * 1);
				} catch (InterruptedException e) {
					break;
				}
				// Try to get the fork on the right
				Object rightLock = listOfLocks[(threadIndex + 1) % numberOfThreads];
				synchronized (rightLock) {
					dinerTable[i + 2] = "<O>";
					dinerTable[i + 3] = "^";
					dinerTable[(i + 4) % (4 * numberOfThreads)] = " ";
					try {
						sleep(unitOfTime * 1);
					} catch (InterruptedException e) {
						break;
					}
					dinerTable[i] = "^";
					dinerTable[i + 1] = " ";
					dinerTable[i + 2] = "_O_";
					dinerTable[i + 3] = " ";
					dinerTable[(i + 4) % (4 * numberOfThreads)] = "^";
				}
			}
		}
	}
	public static void main(String[] a) {
		for (int i = 0; i < numberOfThreads; i++)
			listOfLocks[i] = new Object();
		for (int i = 0; i < numberOfThreads; i++) {
			dinerTable[4 * i] = "^";
			dinerTable[4 * i + 1] = " ";
			dinerTable[4 * i + 2] = "_O_";
			dinerTable[4 * i + 3] = " ";
			lockedDiner[4 * i] = " ";
			lockedDiner[4 * i + 1] = "^";
			lockedDiner[4 * i + 2] = "<O ";
			lockedDiner[4 * i + 3] = " ";
		}
		for (int i = 0; i < numberOfThreads; i++) {
			Thread t = new Philosopher(i);
			t.setDaemon(true);
			t.start();
		}
		String lockedString = array2String(lockedDiner, "");
		System.out.println("The diner table:");
		long step = 0;
		while (true) {
			step++;
			String sDinerTable = array2String(dinerTable, "");
			System.out.println(sDinerTable + "   " + step + "\t" + getInfo(dinerTable));
			if (lockedString.equals(sDinerTable)) {
				break;
			}
			try {
				Thread.sleep(unitOfTime);
			} catch (InterruptedException e) {
				System.out.println("Interrupted.");
			}
		}
		System.out.println("The diner is locked.");
	}
}
